@@ -1,6 +1,6 @@
 """Type checking of attribute access"""
 
-from typing import cast, Callable, List
+from typing import cast, Callable, List, Optional
 
 from mypy.types import (
     Type, Instance, AnyType, TupleType, CallableType, FunctionLike, TypeVarDef,
@@ -187,7 +187,7 @@ def analyze_var(name: str, var: Var, itype: Instance, info: TypeInfo, node: Cont
                 # methods: the former to the instance, the latter to the
                 # class.
                 functype = cast(FunctionLike, t)
-                check_method_type(functype, itype, node, msg)
+                check_method_type(functype, itype, var.is_classmethod, node, msg)
                 signature = method_type(functype)
                 if var.is_property:
                     # A property cannot have an overloaded type => the cast
@@ -228,17 +228,29 @@ def lookup_member_var_or_accessor(info: TypeInfo, name: str,
         return None
 
 
-def check_method_type(functype: FunctionLike, itype: Instance,
+def check_method_type(functype: FunctionLike, itype: Instance, is_classmethod: bool,
                       context: Context, msg: MessageBuilder) -> None:
     for item in functype.items():
-        if not item.arg_types or item.arg_kinds[0] != ARG_POS:
-            # No positional first (self) argument.
+        if not item.arg_types or item.arg_kinds[0] not in (ARG_POS, ARG_STAR):
+            # No positional first (self) argument (*args is okay).
             msg.invalid_method_type(item, context)
-        else:
+        elif not is_classmethod:
             # Check that self argument has type 'Any' or valid instance type.
             selfarg = item.arg_types[0]
             if not subtypes.is_equivalent(selfarg, itype):
                 msg.invalid_method_type(item, context)
+        else:
+            # Check that cls argument has type 'Any' or valid class type.
+            # (This is sufficient for the current treatment of @classmethod,
+            # but probably needs to be revisited when we implement Type[C]
+            # or advanced variants of it like Type[<args>, C].)
+            clsarg = item.arg_types[0]
+            if isinstance(clsarg, CallableType) and clsarg.is_type_obj():
+                if not subtypes.is_equivalent(clsarg.ret_type, itype):
+                    msg.invalid_class_method_type(item, context)
+            else:
+                if not subtypes.is_equivalent(clsarg, AnyType()):
+                    msg.invalid_class_method_type(item, context)
 
 
 def analyze_class_attribute_access(itype: Instance,
@@ -276,7 +288,7 @@ def analyze_class_attribute_access(itype: Instance,
         return AnyType()
 
     if isinstance(node.node, TypeInfo):
-        return type_object_type(cast(TypeInfo, node.node), builtin_type)
+        return type_object_type(node.node, builtin_type)
 
     if is_decorated:
         # TODO: Return type of decorated function. This is quick hack to work around #998.
@@ -338,7 +350,7 @@ def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) ->
                                    arg_names=["_args", "_kwds"],
                                    ret_type=AnyType(),
                                    fallback=builtin_type('builtins.function'))
-                return class_callable(sig, info, fallback)
+                return class_callable(sig, info, fallback, None)
         # Construct callable type based on signature of __init__. Adjust
         # return type and insert type arguments.
         return type_object_type_from_function(init_method, info, fallback)
@@ -347,17 +359,37 @@ def type_object_type(info: TypeInfo, builtin_type: Callable[[str], Instance]) ->
 def type_object_type_from_function(init_or_new: FuncBase, info: TypeInfo,
                                    fallback: Instance) -> FunctionLike:
     signature = method_type_with_fallback(init_or_new, fallback)
+
+    # The __init__ method might come from a generic superclass
+    # (init_or_new.info) with type variables that do not map
+    # identically to the type variables of the class being constructed
+    # (info). For example
+    #
+    #   class A(Generic[T]): def __init__(self, x: T) -> None: pass
+    #   class B(A[List[T]], Generic[T]): pass
+    #
+    # We need to first map B's __init__ to the type (List[T]) -> None.
+    signature = cast(FunctionLike,
+                     map_type_from_supertype(signature, info, init_or_new.info))
+
+    if init_or_new.info.fullname() == 'builtins.dict':
+        # Special signature!
+        special_sig = 'dict'
+    else:
+        special_sig = None
+
     if isinstance(signature, CallableType):
-        return class_callable(signature, info, fallback)
+        return class_callable(signature, info, fallback, special_sig)
     else:
         # Overloaded __init__/__new__.
         items = []  # type: List[CallableType]
         for item in cast(Overloaded, signature).items():
-            items.append(class_callable(item, info, fallback))
+            items.append(class_callable(item, info, fallback, special_sig))
         return Overloaded(items)
 
 
-def class_callable(init_type: CallableType, info: TypeInfo, type_type: Instance) -> CallableType:
+def class_callable(init_type: CallableType, info: TypeInfo, type_type: Instance,
+                   special_sig: Optional[str]) -> CallableType:
     """Create a type object type based on the signature of __init__."""
     variables = []  # type: List[TypeVarDef]
     for i, tvar in enumerate(info.defn.type_vars):
@@ -368,9 +400,12 @@ def class_callable(init_type: CallableType, info: TypeInfo, type_type: Instance)
     variables.extend(initvars)
 
     callable_type = init_type.copy_modified(
-        ret_type=self_type(info), fallback=type_type, name=None, variables=variables)
+        ret_type=self_type(info), fallback=type_type, name=None, variables=variables,
+        special_sig=special_sig)
     c = callable_type.with_name('"{}"'.format(info.name()))
-    return convert_class_tvars_to_func_tvars(c, len(initvars))
+    cc = convert_class_tvars_to_func_tvars(c, len(initvars))
+    cc.is_classmethod_class = True
+    return cc
 
 
 def convert_class_tvars_to_func_tvars(callable: CallableType,
@@ -402,3 +437,33 @@ class TvarTranslator(TypeTranslator):
             else:
                 items.append(v)
         return items
+
+
+def map_type_from_supertype(typ: Type, sub_info: TypeInfo,
+                            super_info: TypeInfo) -> Type:
+    """Map type variables in a type defined in a supertype context to be valid
+    in the subtype context. Assume that the result is unique; if more than
+    one type is possible, return one of the alternatives.
+
+    For example, assume
+
+    . class D(Generic[S]) ...
+    . class C(D[E[T]], Generic[T]) ...
+
+    Now S in the context of D would be mapped to E[T] in the context of C.
+    """
+    # Create the type of self in subtype, of form t[a1, ...].
+    inst_type = self_type(sub_info)
+    if isinstance(inst_type, TupleType):
+        inst_type = inst_type.fallback
+    # Map the type of self to supertype. This gets us a description of the
+    # supertype type variables in terms of subtype variables, i.e. t[t1, ...]
+    # so that any type variables in tN are to be interpreted in subtype
+    # context.
+    inst_type = map_instance_to_supertype(inst_type, super_info)
+    # Finally expand the type variables in type with those in the previously
+    # constructed type. Note that both type and inst_type may have type
+    # variables, but in type they are interpreterd in supertype context while
+    # in inst_type they are interpreted in subtype context. This works even if
+    # the names of type variables in supertype and subtype overlap.
+    return expand_type_by_instance(typ, inst_type)
