@@ -11,9 +11,12 @@ from typing import cast, List, Dict, Any, Sequence, Iterable, Tuple
 from mypy.errors import Errors
 from mypy.types import (
     Type, CallableType, Instance, TypeVarType, TupleType, UnionType, Void, NoneTyp, AnyType,
-    Overloaded, FunctionLike, DeletedType, PartialType
+    Overloaded, FunctionLike, DeletedType
+)
+from mypy.nodes import (
+    TypeInfo, Context, MypyFile, op_methods, FuncDef, reverse_type_aliases,
+    ARG_STAR, ARG_STAR2
 )
-from mypy.nodes import TypeInfo, Context, MypyFile, op_methods, FuncDef, reverse_type_aliases
 
 
 # Constants that represent simple type checker error message, i.e. messages
@@ -69,10 +72,15 @@ READ_ONLY_PROPERTY_OVERRIDES_READ_WRITE = \
 INSTANCE_LAYOUT_CONFLICT = 'Instance layout conflict in multiple inheritance'
 FORMAT_REQUIRES_MAPPING = 'Format requires a mapping'
 GENERIC_TYPE_NOT_VALID_AS_EXPRESSION = \
-    "Generic type not valid as an expression any more (use '# type:' comment instead)"
+    "Generic type is prohibited as a runtime expression (use a type alias or '# type:' comment)"
 RETURN_TYPE_CANNOT_BE_CONTRAVARIANT = "Cannot use a contravariant type variable as return type"
 FUNCTION_PARAMETER_CANNOT_BE_COVARIANT = "Cannot use a covariant type variable as a parameter"
 INCOMPATIBLE_IMPORT_OF = "Incompatible import of"
+FUNCTION_TYPE_EXPECTED = "Function is missing a type annotation"
+RETURN_TYPE_EXPECTED = "Function is missing a return type annotation"
+ARGUMENT_TYPE_EXPECTED = "Function is missing a type annotation for one or more arguments"
+KEYWORD_ARGUMENT_REQUIRES_STR_KEY_TYPE = \
+    'Keyword argument only valid with "str" key type in call to "dict"'
 
 
 class MessageBuilder:
@@ -152,7 +160,7 @@ class MessageBuilder:
             # If format_simple returns a non-trivial result, use that.
             return s
         elif isinstance(typ, FunctionLike):
-            func = cast(FunctionLike, typ)
+            func = typ
             if func.is_type_obj():
                 # The type of a type object type can be derived from the
                 # return type (this always works).
@@ -191,7 +199,7 @@ class MessageBuilder:
           callable type -> "" (empty string)
         """
         if isinstance(typ, Instance):
-            itype = cast(Instance, typ)
+            itype = typ
             # Get the short name of the type.
             if verbosity >= 2:
                 base_str = itype.type.fullname()
@@ -225,13 +233,13 @@ class MessageBuilder:
                     return '{}[...]'.format(base_str)
         elif isinstance(typ, TypeVarType):
             # This is similar to non-generic instance types.
-            return '"{}"'.format((cast(TypeVarType, typ)).name)
+            return '"{}"'.format(typ.name)
         elif isinstance(typ, TupleType):
             # Prefer the name of the fallback class (if not tuple), as it's more informative.
             if typ.fallback.type.fullname() != 'builtins.tuple':
                 return self.format_simple(typ.fallback)
             items = []
-            for t in (cast(TupleType, typ)).items:
+            for t in typ.items:
                 items.append(strip_quotes(self.format(t)))
             s = '"Tuple[{}]"'.format(', '.join(items))
             if len(s) < 40:
@@ -240,7 +248,7 @@ class MessageBuilder:
                 return 'tuple(length {})'.format(len(items))
         elif isinstance(typ, UnionType):
             items = []
-            for t in (cast(UnionType, typ)).items:
+            for t in typ.items:
                 items.append(strip_quotes(self.format(t)))
             s = '"Union[{}]"'.format(', '.join(items))
             if len(s) < 40:
@@ -292,7 +300,7 @@ class MessageBuilder:
         messages. Return type Any.
         """
         if (isinstance(typ, Instance) and
-                (cast(Instance, typ)).type.has_readable_member(member)):
+                typ.type.has_readable_member(member)):
             self.fail('Member "{}" is not assignable'.format(member), context)
         elif isinstance(typ, Void):
             self.check_void(typ, context)
@@ -328,8 +336,7 @@ class MessageBuilder:
             # The non-special case: a missing ordinary attribute.
             if not self.disable_type_names:
                 failed = False
-                if isinstance(typ, Instance) and cast(Instance, typ).type.names:
-                    typ = cast(Instance, typ)
+                if isinstance(typ, Instance) and typ.type.names:
                     alternatives = set(typ.type.names.keys())
                     matches = [m for m in COMMON_MISTAKES.get(member, []) if m in alternatives]
                     matches.extend(best_matches(member, alternatives)[:3])
@@ -391,8 +398,13 @@ class MessageBuilder:
         self.fail('{} not callable'.format(self.format(typ)), context)
         return AnyType()
 
+    def untyped_function_call(self, callee: CallableType, context: Context) -> Type:
+        name = callee.name if callee.name is not None else '(unknown)'
+        self.fail('call to untyped function {} in typed context'.format(name), context)
+        return AnyType()
+
     def incompatible_argument(self, n: int, m: int, callee: CallableType, arg_type: Type,
-                              context: Context) -> None:
+                              arg_kind: int, context: Context) -> None:
         """Report an error about an incompatible argument type.
 
         The argument type is arg_type, argument number is n and the
@@ -457,6 +469,10 @@ class MessageBuilder:
             except IndexError:  # Varargs callees
                 expected_type = callee.arg_types[-1]
             arg_type_str, expected_type_str = self.format_distinctly(arg_type, expected_type)
+            if arg_kind == ARG_STAR:
+                arg_type_str = '*' + arg_type_str
+            elif arg_kind == ARG_STAR2:
+                arg_type_str = '**' + arg_type_str
             msg = 'Argument {} {}has incompatible type {}; expected {}'.format(
                 n, target, arg_type_str, expected_type_str)
         self.fail(msg, context)
@@ -664,8 +680,7 @@ class MessageBuilder:
         self.fail('List or tuple expected as variable arguments', context)
 
     def invalid_keyword_var_arg(self, typ: Type, context: Context) -> None:
-        if isinstance(typ, Instance) and (
-                (cast(Instance, typ)).type.fullname() == 'builtins.dict'):
+        if isinstance(typ, Instance) and (typ.type.fullname() == 'builtins.dict'):
             self.fail('Keywords must be strings', context)
         else:
             self.fail('Argument after ** must be a dictionary',
@@ -725,6 +740,9 @@ class MessageBuilder:
     def invalid_method_type(self, sig: CallableType, context: Context) -> None:
         self.fail('Invalid method type', context)
 
+    def invalid_class_method_type(self, sig: CallableType, context: Context) -> None:
+        self.fail('Invalid class method type', context)
+
     def incompatible_conditional_function_def(self, defn: FuncDef) -> None:
         self.fail('All conditional function variants must have identical '
                   'signatures', defn)
