@@ -20,11 +20,11 @@ from mypy.nodes import (
     ClassDef, Decorator, Block, Var, OperatorAssignmentStmt,
     ExpressionStmt, AssignmentStmt, ReturnStmt, RaiseStmt, AssertStmt,
     DelStmt, BreakStmt, ContinueStmt, PassStmt, GlobalDecl,
-    WhileStmt, ForStmt, IfStmt, TryStmt, WithStmt, CastExpr,
+    WhileStmt, ForStmt, IfStmt, TryStmt, WithStmt,
     TupleExpr, GeneratorExpr, ListComprehension, ListExpr, ConditionalExpr,
     DictExpr, SetExpr, NameExpr, IntExpr, StrExpr, BytesExpr, UnicodeExpr,
     FloatExpr, CallExpr, SuperExpr, MemberExpr, IndexExpr, SliceExpr, OpExpr,
-    UnaryExpr, FuncExpr, TypeApplication, PrintStmt, ImportBase, ComparisonExpr,
+    UnaryExpr, FuncExpr, PrintStmt, ImportBase, ComparisonExpr,
     StarExpr, YieldFromExpr, NonlocalDecl, DictionaryComprehension,
     SetComprehension, ComplexExpr, EllipsisExpr, YieldExpr, ExecStmt, Argument,
     BackquoteExpr
@@ -32,7 +32,7 @@ from mypy.nodes import (
 from mypy import defaults
 from mypy import nodes
 from mypy.errors import Errors, CompileError
-from mypy.types import Void, Type, CallableType, AnyType, UnboundType
+from mypy.types import Type, CallableType, AnyType, UnboundType
 from mypy.parsetype import (
     parse_type, parse_types, parse_signature, TypeParseError, parse_str_as_signature
 )
@@ -70,7 +70,8 @@ none = Token('')  # Empty token
 
 def parse(source: Union[str, bytes], fnam: str = None, errors: Errors = None,
           pyversion: Tuple[int, int] = defaults.PYTHON3_VERSION,
-          custom_typing_module: str = None, implicit_any: bool = False) -> MypyFile:
+          custom_typing_module: str = None,
+          fast_parser: bool = False) -> MypyFile:
     """Parse a source file, without doing any semantic analysis.
 
     Return the parse tree. If errors is not provided, raise ParseError
@@ -78,13 +79,20 @@ def parse(source: Union[str, bytes], fnam: str = None, errors: Errors = None,
 
     The pyversion (major, minor) argument determines the Python syntax variant.
     """
+    if fast_parser:
+        import mypy.fastparse
+        return mypy.fastparse.parse(source,
+                                    fnam=fnam,
+                                    errors=errors,
+                                    pyversion=pyversion,
+                                    custom_typing_module=custom_typing_module)
+
     is_stub_file = bool(fnam) and fnam.endswith('.pyi')
     parser = Parser(fnam,
                     errors,
                     pyversion,
                     custom_typing_module,
-                    is_stub_file=is_stub_file,
-                    implicit_any=implicit_any)
+                    is_stub_file=is_stub_file)
     tree = parser.parse(source)
     tree.path = fnam
     tree.is_stub = is_stub_file
@@ -116,13 +124,11 @@ class Parser:
     ignored_lines = None  # type: Set[int]
 
     def __init__(self, fnam: str, errors: Errors, pyversion: Tuple[int, int],
-                 custom_typing_module: str = None, is_stub_file: bool = False,
-                 implicit_any = False) -> None:
+                 custom_typing_module: str = None, is_stub_file: bool = False) -> None:
         self.raise_on_error = errors is None
         self.pyversion = pyversion
         self.custom_typing_module = custom_typing_module
         self.is_stub_file = is_stub_file
-        self.implicit_any = implicit_any
         if errors is not None:
             self.errors = errors
         else:
@@ -329,7 +335,7 @@ class Parser:
         metaclass = None
 
         try:
-            commas, base_types = [], []  # type: List[Token], List[Node]
+            base_types = []  # type: List[Node]
             try:
                 name_tok = self.expect_type(Name)
                 name = name_tok.string
@@ -341,13 +347,13 @@ class Parser:
                     while True:
                         if self.current_str() == ')':
                             break
-                        if self.current_str() == 'metaclass':
-                            metaclass = self.parse_metaclass()
+                        if self.peek().string == '=':
+                            metaclass = self.parse_class_keywords()
                             break
                         base_types.append(self.parse_super_type())
                         if self.current_str() != ',':
                             break
-                        commas.append(self.skip())
+                        self.skip()
                     self.expect(')')
             except ParseError:
                 pass
@@ -360,14 +366,29 @@ class Parser:
             self.errors.pop_type()
             self.is_class_body = old_is_class_body
 
+    def parse_class_keywords(self) -> Optional[str]:
+        """Parse the class keyword arguments, keeping the metaclass but
+        ignoring all others.  Returns None if the metaclass isn't found.
+        """
+        metaclass = None
+        while True:
+            key = self.expect_type(Name)
+            self.expect('=')
+            if key.string == 'metaclass':
+                metaclass = self.parse_qualified_name()
+            else:
+                # skip the class value
+                self.parse_expression(precedence[','])
+            if self.current_str() != ',':
+                break
+            self.skip()
+            if self.current_str() == ')':
+                break
+        return metaclass
+
     def parse_super_type(self) -> Node:
         return self.parse_expression(precedence[','])
 
-    def parse_metaclass(self) -> str:
-        self.expect('metaclass')
-        self.expect('=')
-        return self.parse_qualified_name()
-
     def parse_decorated_function_or_class(self) -> Node:
         decorators = []
         no_type_checks = False
@@ -421,7 +442,18 @@ class Parser:
                     self.errors.report(
                         def_tok.line, 'Function has duplicate type signatures')
                 sig = cast(CallableType, comment_type)
-                if is_method and len(sig.arg_kinds) < len(arg_kinds):
+                if sig.is_ellipsis_args:
+                    # When we encounter an ellipsis, fill in the arg_types with
+                    # a bunch of AnyTypes, emulating Callable[..., T]
+                    arg_types = [AnyType()] * len(arg_kinds)  # type: List[Type]
+                    typ = CallableType(
+                        arg_types,
+                        arg_kinds,
+                        arg_names,
+                        sig.ret_type,
+                        None,
+                        line=def_tok.line)
+                elif is_method and len(sig.arg_kinds) < len(arg_kinds):
                     self.check_argument_kinds(arg_kinds,
                                               [nodes.ARG_POS] + sig.arg_kinds,
                                               def_tok.line)
@@ -450,20 +482,6 @@ class Parser:
             if is_error:
                 return None
 
-            # add implicit anys
-            if typ is None and self.implicit_any and not self.is_stub_file:
-                ret_type = None  # type: Type
-                if is_method and name == '__init__':
-                    ret_type = UnboundType('None', [])
-                else:
-                    ret_type = AnyType()
-                typ = CallableType([AnyType() for _ in args],
-                                   arg_kinds,
-                                   [a.variable.name() for a in args],
-                                   ret_type,
-                                   None,
-                                   line=def_tok.line)
-
             node = FuncDef(name, args, body, typ)
             node.set_line(def_tok)
             if typ is not None:
@@ -628,9 +646,10 @@ class Parser:
                 elif self.current_str() == '(':
                     arg, extra_stmt, names = self.parse_tuple_arg(len(args))
                     args.append(arg)
-                    extra_stmts.append(extra_stmt)
+                    if extra_stmt is not None:
+                        extra_stmts.append(extra_stmt)
+                        has_tuple_arg = True
                     arg_names.extend(names)
-                    has_tuple_arg = True
                 else:
                     arg, require_named = self.parse_normal_arg(
                         require_named,
@@ -688,6 +707,10 @@ class Parser:
         function body (the second return value).
 
         Return tuple (argument, decomposing assignment, list of names defined).
+
+        Special case: if the argument is just (x) then it's not a tuple;
+        we indicate this by returning (argument, None, ['x']).
+        However, if the argument is (x,) then it *is* a (singleton) tuple.
         """
         line = self.current().line
         # Generate a new argument name that is very unlikely to clash with anything.
@@ -697,13 +720,14 @@ class Parser:
         paren_arg = self.parse_parentheses()
         self.verify_tuple_arg(paren_arg)
         if isinstance(paren_arg, NameExpr):
-            # This isn't a tuple. Revert to a normal argument. We'll still get a no-op
-            # assignment below but that's benign.
+            # This isn't a tuple. Revert to a normal argument.
             arg_name = paren_arg.name
-        rvalue = NameExpr(arg_name)
-        rvalue.set_line(line)
-        decompose = AssignmentStmt([paren_arg], rvalue)
-        decompose.set_line(line)
+            decompose = None
+        else:
+            rvalue = NameExpr(arg_name)
+            rvalue.set_line(line)
+            decompose = AssignmentStmt([paren_arg], rvalue)
+            decompose.set_line(line)
         kind = nodes.ARG_POS
         initializer = None
         if self.current_str() == '=':
@@ -791,13 +815,13 @@ class Parser:
         arg_types = [arg.type_annotation for arg in args]
         for i in range(len(arg_types)):
             if arg_types[i] is None:
-                arg_types[i] = AnyType()
+                arg_types[i] = AnyType(implicit=True)
         if ret_type is None:
-            ret_type = AnyType()
+            ret_type = AnyType(implicit=True)
         arg_kinds = [arg.kind for arg in args]
         arg_names = [arg.variable.name() for arg in args]
         return CallableType(arg_types, arg_kinds, arg_names, ret_type, None, name=None,
-                        variables=None, bound_vars=[], line=line)
+                        variables=None, line=line)
 
     # Parsing statements
 
@@ -1837,7 +1861,7 @@ class Parser:
         self.parse_error_at(self.current())
         raise ParseError()
 
-    def parse_error_at(self, tok: Token, skip: bool = True) -> None:
+    def parse_error_at(self, tok: Token, skip: bool = True, reason: Optional[str] = None) -> None:
         msg = ''
         if isinstance(tok, LexError):
             msg = token_repr(tok)
@@ -1845,7 +1869,8 @@ class Parser:
         elif isinstance(tok, Indent) or isinstance(tok, Dedent):
             msg = 'Inconsistent indentation'
         else:
-            msg = 'Parse error before {}'.format(token_repr(tok))
+            formatted_reason = ": {}".format(reason) if reason else ""
+            msg = 'Parse error before {}{}'.format(token_repr(tok), formatted_reason)
 
         self.errors.report(tok.line, msg)
 
@@ -1878,7 +1903,7 @@ class Parser:
         try:
             typ, self.ind = parse_type(self.tok, self.ind)
         except TypeParseError as e:
-            self.parse_error_at(e.token)
+            self.parse_error_at(e.token, reason=e.message)
             raise ParseError()
         return typ
 
@@ -1908,7 +1933,7 @@ class Parser:
                 else:
                     type, index = parse_signature(tokens)
             except TypeParseError as e:
-                self.parse_error_at(e.token, skip=False)
+                self.parse_error_at(e.token, skip=False, reason=e.message)
                 return None
             if index < len(tokens) - 2:
                 self.parse_error_at(tokens[index], skip=False)
@@ -1968,7 +1993,7 @@ if __name__ == '__main__':
     import sys
 
     def usage():
-        print('Usage: parse.py [--py2] [--quiet] FILE [...]')
+        print('Usage: parse.py [--py2] [--quiet] FILE [...]', file=sys.stderr)
         sys.exit(2)
 
     args = sys.argv[1:]
