@@ -1,7 +1,7 @@
 """Classes for representing mypy types."""
 
 from abc import abstractmethod
-from typing import Any, TypeVar, List, Tuple, cast, Generic, Set, Sequence, Optional
+from typing import Any, TypeVar, Dict, List, Tuple, cast, Generic, Set, Sequence, Optional
 
 import mypy.nodes
 from mypy.nodes import INVARIANT, SymbolNode
@@ -9,6 +9,8 @@ from mypy.nodes import INVARIANT, SymbolNode
 
 T = TypeVar('T')
 
+JsonDict = Dict[str, Any]
+
 
 class Type(mypy.nodes.Context):
     """Abstract base class for all types."""
@@ -27,18 +29,31 @@ class Type(mypy.nodes.Context):
     def __repr__(self) -> str:
         return self.accept(TypeStrVisitor())
 
+    def serialize(self) -> JsonDict:
+        raise NotImplementedError('Cannot serialize {} instance'.format(self.__class__.__name__))
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'Type':
+        classname = data['.class']
+        glo = globals()
+        if classname in glo:
+            cl = glo[classname]
+            if 'deserialize' in cl.__dict__:
+                return cl.deserialize(data)
+        raise NotImplementedError('unexpected .class {}'.format(classname))
+
 
 class TypeVarDef(mypy.nodes.Context):
     """Definition of a single type variable."""
 
     name = ''
     id = 0
-    values = None  # type: List[Type]
+    values = None  # type: Optional[List[Type]]
     upper_bound = None  # type: Type
     variance = INVARIANT  # type: int
     line = 0
 
-    def __init__(self, name: str, id: int, values: List[Type],
+    def __init__(self, name: str, id: int, values: Optional[List[Type]],
                  upper_bound: Type, variance: int = INVARIANT, line: int = -1) -> None:
         self.name = name
         self.id = id
@@ -53,9 +68,31 @@ class TypeVarDef(mypy.nodes.Context):
     def __repr__(self) -> str:
         if self.values:
             return '{} in {}'.format(self.name, tuple(self.values))
+        elif not is_named_instance(self.upper_bound, 'builtins.object'):
+            return '{} <: {}'.format(self.name, self.upper_bound)
         else:
             return self.name
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'TypeVarDef',
+                'name': self.name,
+                'id': self.id,
+                'values': None if self.values is None else [v.serialize() for v in self.values],
+                'upper_bound': self.upper_bound.serialize(),
+                'variance': self.variance,
+                }
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'TypeVarDef':
+        assert data['.class'] == 'TypeVarDef'
+        return TypeVarDef(data['name'],
+                          data['id'],
+                          None if data['values'] is None
+                          else [Type.deserialize(v) for v in data['values']],
+                          Type.deserialize(data['upper_bound']),
+                          data['variance'],
+                          )
+
 
 class UnboundType(Type):
     """Instance type that has not been bound during semantic analysis."""
@@ -73,6 +110,18 @@ class UnboundType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_unbound_type(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'UnboundType',
+                'name': self.name,
+                'args': [a.serialize() for a in self.args],
+                }
+
+    @classmethod
+    def deserialize(self, data: JsonDict) -> 'UnboundType':
+        assert data['.class'] == 'UnboundType'
+        return UnboundType(data['name'],
+                           [Type.deserialize(a) for a in data['args']])
+
 
 class ErrorType(Type):
     """The error type is used as the result of failed type operations."""
@@ -98,13 +147,35 @@ class TypeList(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_type_list(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'TypeList',
+                'items': [t.serialize() for t in self.items],
+                }
+
+    @classmethod
+    def deserialize(self, data: JsonDict) -> 'TypeList':
+        assert data['.class'] == 'TypeList'
+        return TypeList([Type.deserialize(t) for t in data['items']])
+
 
 class AnyType(Type):
     """The type 'Any'."""
 
+    def __init__(self, implicit=False, line: int = -1) -> None:
+        super().__init__(line)
+        self.implicit = implicit
+
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_any(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'AnyType'}
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'AnyType':
+        assert data['.class'] == 'AnyType'
+        return AnyType()
+
 
 class Void(Type):
     """The return type 'None'.
@@ -125,6 +196,14 @@ class Void(Type):
     def with_source(self, source: str) -> 'Void':
         return Void(source, self.line)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'Void'}
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'Void':
+        assert data['.class'] == 'Void'
+        return Void()
+
 
 class NoneTyp(Type):
     """The type of 'None'.
@@ -145,6 +224,14 @@ class NoneTyp(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_none_type(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'NoneTyp'}
+
+    @classmethod
+    def deserialize(self, data: JsonDict) -> 'NoneTyp':
+        assert data['.class'] == 'NoneTyp'
+        return NoneTyp()
+
 
 class ErasedType(Type):
     """Placeholder for an erased type.
@@ -172,6 +259,15 @@ class DeletedType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_deleted_type(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'DeletedType',
+                'source': self.source}
+
+    @classmethod
+    def deserialize(self, data: JsonDict) -> 'DeletedType':
+        assert data['.class'] == 'DeletedType'
+        return DeletedType(data['source'])
+
 
 class Instance(Type):
     """An instance type of form C[T1, ..., Tn].
@@ -193,6 +289,29 @@ class Instance(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_instance(self)
 
+    type_ref = None  # type: str
+
+    def serialize(self) -> JsonDict:
+        data = {'.class': 'Instance',
+                }  # type: JsonDict
+        assert self.type is not None
+        data['type_ref'] = self.type.alt_fullname or self.type.fullname()
+        if self.args:
+            data['args'] = [arg.serialize() for arg in self.args]
+        return data
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'Instance':
+        assert data['.class'] == 'Instance'
+        args = []  # type: List[Type]
+        if 'args' in data:
+            args_list = data['args']
+            assert isinstance(args_list, list)
+            args = [Type.deserialize(arg) for arg in args_list]
+        inst = Instance(None, args)
+        inst.type_ref = data['type_ref']  # Will be fixed up by fixup.py later.
+        return inst
+
 
 class TypeVarType(Type):
     """A type variable type.
@@ -204,7 +323,7 @@ class TypeVarType(Type):
     name = ''  # Name of the type variable (for messages and debugging)
     id = 0     # 1, 2, ... for type-related, -1, ... for function-related
     values = None  # type: List[Type]  # Value restriction, empty list if no restriction
-    upper_bound = None  # type: Type   # Upper bound for values (currently always 'object')
+    upper_bound = None  # type: Type   # Upper bound for values
     # See comments in TypeVarDef for more about variance.
     variance = INVARIANT  # type: int
 
@@ -220,6 +339,30 @@ class TypeVarType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_type_var(self)
 
+    def erase_to_union_or_bound(self) -> Type:
+        if self.values:
+            return UnionType.make_simplified_union(self.values)
+        else:
+            return self.upper_bound
+
+    def serialize(self) -> JsonDict:
+        return {'.class': 'TypeVarType',
+                'name': self.name,
+                'id': self.id,
+                'values': [v.serialize() for v in self.values],
+                'upper_bound': self.upper_bound.serialize(),
+                'variance': self.variance,
+                }
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'TypeVarType':
+        assert data['.class'] == 'TypeVarType'
+        return TypeVarType(data['name'],
+                           data['id'],
+                           [Type.deserialize(v) for v in data['values']],
+                           Type.deserialize(data['upper_bound']),
+                           data['variance'])
+
 
 class FunctionLike(Type):
     """Abstract base class for function types."""
@@ -227,6 +370,9 @@ class FunctionLike(Type):
     @abstractmethod
     def is_type_obj(self) -> bool: pass
 
+    def is_concrete_type_obj(self) -> bool:
+        return self.is_type_obj()
+
     @abstractmethod
     def type_object(self) -> mypy.nodes.TypeInfo: pass
 
@@ -239,6 +385,10 @@ class FunctionLike(Type):
     # Corresponding instance type (e.g. builtins.type)
     fallback = None  # type: Instance
 
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'FunctionLike':
+        return cast(FunctionLike, super().deserialize(data))
+
 
 _dummy = object()  # type: Any
 
@@ -249,33 +399,26 @@ class CallableType(FunctionLike):
     arg_types = None  # type: List[Type]  # Types of function arguments
     arg_kinds = None  # type: List[int]   # mypy.nodes.ARG_ constants
     arg_names = None  # type: List[str]   # None if not a keyword argument
-    min_args = 0                    # Minimum number of arguments
-    is_var_arg = False              # Is it a varargs function?
-    ret_type = None  # type:Type    # Return value type
+    min_args = 0                    # Minimum number of arguments; derived from arg_kinds
+    is_var_arg = False              # Is it a varargs function?  derived from arg_kinds
+    ret_type = None  # type: Type   # Return value type
     name = ''                       # Name (may be None; for error messages)
     definition = None  # type: SymbolNode # For error messages.  May be None.
     # Type variables for a generic function
     variables = None  # type: List[TypeVarDef]
 
-    # Implicit bound values of type variables. These can be either for
-    # class type variables or for generic function type variables.
-    # For example, the method 'append' of List[int] has implicit value
-    # 'int' for the list type variable; the explicit method type is
-    # just 'def append(int) -> None', without any type variable. Implicit
-    # values are needed for runtime type checking, but they do not
-    # affect static type checking.
-    #
-    # All class type arguments must be stored first, ordered by id,
-    # and function type arguments must be stored next, again ordered by id
-    # (absolute value this time).
-    #
-    # Stored as tuples (id, type).
-    bound_vars = None  # type: List[Tuple[int, Type]]
-
     # Is this Callable[..., t] (with literal '...')?
     is_ellipsis_args = False
+    # Is this callable constructed for the benefit of a classmethod's 'cls' argument?
+    is_classmethod_class = False
+    # Was this type implicitly generated instead of explicitly specified by the user?
+    implicit = False
+    # Defined for signatures that require special handling (currently only value is 'dict'
+    # for a signature similar to 'dict')
+    special_sig = None  # type: Optional[str]
 
-    def __init__(self, arg_types: List[Type],
+    def __init__(self,
+                 arg_types: List[Type],
                  arg_kinds: List[int],
                  arg_names: List[str],
                  ret_type: Type,
@@ -283,13 +426,14 @@ class CallableType(FunctionLike):
                  name: str = None,
                  definition: SymbolNode = None,
                  variables: List[TypeVarDef] = None,
-                 bound_vars: List[Tuple[int, Type]] = None,
                  line: int = -1,
-                 is_ellipsis_args: bool = False) -> None:
+                 is_ellipsis_args: bool = False,
+                 implicit=False,
+                 is_classmethod_class=False,
+                 special_sig=None,
+                 ) -> None:
         if variables is None:
             variables = []
-        if not bound_vars:
-            bound_vars = []
         self.arg_types = arg_types
         self.arg_kinds = arg_kinds
         self.arg_names = arg_names
@@ -301,8 +445,9 @@ class CallableType(FunctionLike):
         self.name = name
         self.definition = definition
         self.variables = variables
-        self.bound_vars = bound_vars
         self.is_ellipsis_args = is_ellipsis_args
+        self.implicit = implicit
+        self.special_sig = special_sig
         super().__init__(line)
 
     def copy_modified(self,
@@ -314,9 +459,9 @@ class CallableType(FunctionLike):
                       name: str = _dummy,
                       definition: SymbolNode = _dummy,
                       variables: List[TypeVarDef] = _dummy,
-                      bound_vars: List[Tuple[int, Type]] = _dummy,
                       line: int = _dummy,
-                      is_ellipsis_args: bool = _dummy) -> 'CallableType':
+                      is_ellipsis_args: bool = _dummy,
+                      special_sig: Optional[str] = _dummy) -> 'CallableType':
         return CallableType(
             arg_types=arg_types if arg_types is not _dummy else self.arg_types,
             arg_kinds=arg_kinds if arg_kinds is not _dummy else self.arg_kinds,
@@ -326,14 +471,19 @@ class CallableType(FunctionLike):
             name=name if name is not _dummy else self.name,
             definition=definition if definition is not _dummy else self.definition,
             variables=variables if variables is not _dummy else self.variables,
-            bound_vars=bound_vars if bound_vars is not _dummy else self.bound_vars,
             line=line if line is not _dummy else self.line,
             is_ellipsis_args=(
                 is_ellipsis_args if is_ellipsis_args is not _dummy else self.is_ellipsis_args),
+            implicit=self.implicit,
+            is_classmethod_class=self.is_classmethod_class,
+            special_sig=special_sig if special_sig is not _dummy else self.special_sig,
         )
 
     def is_type_obj(self) -> bool:
-        return self.fallback.type.fullname() == 'builtins.type'
+        return self.fallback.type is not None and self.fallback.type.fullname() == 'builtins.type'
+
+    def is_concrete_type_obj(self) -> bool:
+        return self.is_type_obj() and self.is_classmethod_class
 
     def type_object(self) -> mypy.nodes.TypeInfo:
         assert self.is_type_obj()
@@ -370,6 +520,41 @@ class CallableType(FunctionLike):
             a.append(tv.id)
         return a
 
+    def serialize(self) -> JsonDict:
+        # TODO: As an optimization, leave out everything related to
+        # generic functions for non-generic functions.
+        return {'.class': 'CallableType',
+                'arg_types': [(None if t is None else t.serialize())
+                              for t in self.arg_types],
+                'arg_kinds': self.arg_kinds,
+                'arg_names': self.arg_names,
+                'ret_type': self.ret_type.serialize(),
+                'fallback': self.fallback.serialize(),
+                'name': self.name,
+                # We don't serialize the definition (only used for error messages).
+                'variables': [v.serialize() for v in self.variables],
+                'is_ellipsis_args': self.is_ellipsis_args,
+                'implicit': self.implicit,
+                'is_classmethod_class': self.is_classmethod_class,
+                }
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'CallableType':
+        assert data['.class'] == 'CallableType'
+        # TODO: Set definition to the containing SymbolNode?
+        return CallableType([(None if t is None else Type.deserialize(t))
+                             for t in data['arg_types']],
+                            data['arg_kinds'],
+                            data['arg_names'],
+                            Type.deserialize(data['ret_type']),
+                            Instance.deserialize(data['fallback']),
+                            name=data['name'],
+                            variables=[TypeVarDef.deserialize(v) for v in data['variables']],
+                            is_ellipsis_args=data['is_ellipsis_args'],
+                            implicit=data['implicit'],
+                            is_classmethod_class=data['is_classmethod_class'],
+                            )
+
 
 class Overloaded(FunctionLike):
     """Overloaded function type T1, ... Tn, where each Ti is CallableType.
@@ -412,6 +597,16 @@ class Overloaded(FunctionLike):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_overloaded(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'Overloaded',
+                'items': [t.serialize() for t in self.items()],
+                }
+
+    @classmethod
+    def deserialize(self, data: JsonDict) -> 'Overloaded':
+        assert data['.class'] == 'Overloaded'
+        return Overloaded([CallableType.deserialize(t) for t in data['items']])
+
 
 class TupleType(Type):
     """The tuple type Tuple[T1, ..., Tn] (at least one type argument).
@@ -441,6 +636,20 @@ class TupleType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_tuple_type(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'TupleType',
+                'items': [t.serialize() for t in self.items],
+                'fallback': self.fallback.serialize(),
+                'implicit': self.implicit,
+                }
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'TupleType':
+        assert data['.class'] == 'TupleType'
+        return TupleType([Type.deserialize(t) for t in data['items']],
+                         Instance.deserialize(data['fallback']),
+                         implicit=data['implicit'])
+
 
 class StarType(Type):
     """The star type *type_parameter.
@@ -512,10 +721,20 @@ class UnionType(Type):
         TODO: Deal with attributes of TupleType etc.
         TODO: This should probably be refactored to go elsewhere.
         """
-        return all((isinstance(x, UnionType) and cast(UnionType, x).has_readable_member(name)) or
-                   (isinstance(x, Instance) and cast(Instance, x).type.has_readable_member(name))
+        return all((isinstance(x, UnionType) and x.has_readable_member(name)) or
+                   (isinstance(x, Instance) and x.type.has_readable_member(name))
                    for x in self.items)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'UnionType',
+                'items': [t.serialize() for t in self.items],
+                }
+
+    @classmethod
+    def deserialize(cls, data: JsonDict) -> 'UnionType':
+        assert data['.class'] == 'UnionType'
+        return UnionType([Type.deserialize(t) for t in data['items']])
+
 
 class PartialType(Type):
     """Type such as List[?] where type arguments are unknown, or partial None type.
@@ -555,6 +774,14 @@ class EllipsisType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_ellipsis_type(self)
 
+    def serialize(self) -> JsonDict:
+        return {'.class': 'EllipsisType'}
+
+    @classmethod
+    def deserialize(self, data: JsonDict) -> 'EllipsisType':
+        assert data['.class'] == 'EllipsisType'
+        return EllipsisType()
+
 
 #
 # Visitor-related classes
@@ -679,8 +906,7 @@ class TypeTranslator(TypeVisitor[Type]):
     def visit_callable_type(self, t: CallableType) -> Type:
         return t.copy_modified(arg_types=self.translate_types(t.arg_types),
                                ret_type=t.ret_type.accept(self),
-                               variables=self.translate_variables(t.variables),
-                               bound_vars=self.translate_bound_vars(t.bound_vars))
+                               variables=self.translate_variables(t.variables))
 
     def visit_tuple_type(self, t: TupleType) -> Type:
         return TupleType(self.translate_types(t.items),
@@ -699,10 +925,6 @@ class TypeTranslator(TypeVisitor[Type]):
     def translate_types(self, types: List[Type]) -> List[Type]:
         return [t.accept(self) for t in types]
 
-    def translate_bound_vars(
-            self, types: List[Tuple[int, Type]]) -> List[Tuple[int, Type]]:
-        return [(id, t.accept(self)) for id, t in types]
-
     def translate_variables(self,
                             variables: List[TypeVarDef]) -> List[TypeVarDef]:
         return variables
@@ -726,7 +948,6 @@ class TypeStrVisitor(TypeVisitor[str]):
     Do not preserve original formatting.
 
     Notes:
-     - Include implicit bound type variables of callables.
      - Represent unbound types as Foo? or Foo?[...].
      - Represent the NoneTyp type as None.
     """
@@ -763,7 +984,7 @@ class TypeStrVisitor(TypeVisitor[str]):
             return "<Deleted '{}'>".format(t.source)
 
     def visit_instance(self, t):
-        s = t.type.fullname()
+        s = t.type.fullname() if t.type is not None else '<?>'
         if t.erased:
             s += '*'
         if t.args != []:
@@ -805,13 +1026,6 @@ class TypeStrVisitor(TypeVisitor[str]):
         if t.variables:
             s = '{} {}'.format(t.variables, s)
 
-        if t.bound_vars != []:
-            # Include implicit bound type variables.
-            a = []
-            for i, bt in t.bound_vars:
-                a.append('{}:{}'.format(i, bt))
-            s = '[{}] {}'.format(', '.join(a), s)
-
         return 'def {}'.format(s)
 
     def visit_overloaded(self, t):
@@ -822,7 +1036,7 @@ class TypeStrVisitor(TypeVisitor[str]):
 
     def visit_tuple_type(self, t):
         s = self.list_str(t.items)
-        if t.fallback:
+        if t.fallback and t.fallback.type:
             fallback_name = t.fallback.type.fullname()
             if fallback_name != 'builtins.tuple':
                 return 'Tuple[{}, fallback={}]'.format(s, t.fallback.accept(self))
@@ -983,4 +1197,5 @@ def replace_leading_arg_type(t: CallableType, self_type: Type) -> CallableType:
 
 def is_named_instance(t: Type, fullname: str) -> bool:
     return (isinstance(t, Instance) and
-            cast(Instance, t).type.fullname() == fullname)
+            t.type is not None and
+            t.type.fullname() == fullname)
